from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.status import HTTP_400_BAD_REQUEST, HTTP_422_UNPROCESSABLE_ENTITY


async def tableau_http_exception_handler(
    request: Request, exc: HTTPException
) -> JSONResponse:
    headers = getattr(exc, "headers", None)
    if headers:
        return JSONResponse(
            status_code=exc.status_code,
            content={
                "message": f"Server Error: {type(exc).__name__}",
                "info": str(exc.detail),
            },
            headers=headers,
        )
    else:
        return JSONResponse(
            status_code=exc.status_code,
            content={
                "message": f"Server Error: {type(exc).__name__}",
                "info": str(exc.detail),
            },
        )


async def tableau_request_validation_exception_handler(
    request: Request, exc: RequestValidationError
) -> JSONResponse:
    return JSONResponse(
        status_code=HTTP_422_UNPROCESSABLE_ENTITY,
        content={
            "message": f"Server Error: {type(exc).__name__}",
            "info": str(exc),
        },
    )


async def tableau_general_exception_handler(
    request: Request, exc: Exception
) -> JSONResponse:
    return JSONResponse(
        status_code=HTTP_400_BAD_REQUEST,
        content={
            "message": f"Server Error: {type(exc).__name__}",
            "info": str(exc),
        },
    )



import json
from typing import Any, Callable, Dict, MutableMapping, Optional

from fastapi import Request, Response
from fastapi.dependencies.utils import request_body_to_args
from fastapi.exceptions import RequestValidationError
from fastapi.openapi.utils import get_flat_models_from_routes
from fastapi.routing import APIRoute
from pydantic.error_wrappers import ErrorWrapper
from pydantic.schema import field_schema, get_model_name_map

from fastapitableau.logger import logger
from fastapitableau.utils import event_from_receive, remove_prefix, replace_dict_keys


class TableauRoute(APIRoute):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._body_schema: Optional[Dict[str, Any]] = None

    def get_body_field_schema(self):
        # schema, definitions, nested models
        route_flat_models = get_flat_models_from_routes([self])
        route_model_name_map = get_model_name_map(route_flat_models)
        body_field_schema = field_schema(
            self.body_field, model_name_map=route_model_name_map
        )
        return body_field_schema

    @property
    def body_schema(self):
        if not self._body_schema:
            f_schema, f_definitions, f_nested_models = self.get_body_field_schema()

            if "$ref" in f_schema.keys():
                ref = remove_prefix(f_schema["$ref"], "#/definitions/")
                definition = f_definitions[ref]
            else:
                definition = f_schema
        else:
            definition = self._body_schema
        return definition

    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            event = await event_from_receive(request.receive)
            try:
                body = json.loads(event["body"])
            except Exception as e:
                logger.warning(
                    "Failed to parse event body as JSON: %s",
                    event["body"],
                    extra={"scope": request.scope},
                )
                raise e

            _body = await self.ensure_request_body(body, scope=request.scope)
            event["body"] = bytes(json.dumps(_body), encoding="utf-8")

            async def _receive():
                return event

            _request = Request(request.scope, _receive)

            return await original_route_handler(_request)

        return custom_route_handler

    async def ensure_request_body(
        self, body: Dict, scope: MutableMapping = None
    ) -> Dict:
        # Here, we only want to operate on Tableau requests. We have a few options.
        # 1. Duck typing, which is what the main branch does right now. It checks to see if this a dict and contains the keys "script" and "data".
        # 2. Validate. Try to process the body with FastAPI's args and see if it works. If it doesn't, try this method.
        # 3. Use headers.
        body_will_validate = await self.body_will_validate(body)
        if body_will_validate:
            _body = body
            logger.debug(
                "Not rewriting body for request to '%s': %s",
                self.path,
                _body,
                extra={"scope": scope},
            )
        elif isinstance(body, Dict) and set(body.keys()) == {"script", "data"}:
            # It looks like a Tableau request.
            data = body["data"]
            if self.body_schema["type"] == "array":
                # We can handle a single arg from Tableau, raising it up a level.
                # Receiving multiple body params causes an error.
                if len(data) == 1:
                    _body = list(data.values())[0]
                else:
                    error = ValueError(
                        f"The endpoint {self.name} expects to receive a list, but received a {type(data)} from Tableau."
                    )
                    raise RequestValidationError([ErrorWrapper(error, "body")])

            elif self.body_schema["type"] == "object":
                # We'll rename Tableau _arg[n] arguments to the argument names we expect.
                # We order based on argument numbering from Tableau.
                properties = self.body_schema["properties"]
                if len(properties) == len(data):
                    expected_names = list(properties.keys())
                    name_map: Dict[str, str] = {}
                    for i, new_name in enumerate(expected_names):
                        old_name = "_arg" + str(i + 1)
                        name_map[old_name] = new_name
                    _body = replace_dict_keys(data, name_map)
                else:
                    error = ValueError(
                        f"The route {self.name} expects {len(properties)} arguments, but received {len(data)} from Tableau."
                    )
                    raise RequestValidationError([ErrorWrapper(error, "body")])
            logger.debug(
                "Rewriting body for Tableau request to '%s': %s",
                self.path,
                _body,
                extra={"scope": scope},
            )
        return _body

    async def body_will_validate(self, body: Dict[str, Any]):
        """
        This function tries to process the request body using FastAPI's own function, and returns False if it fails to process. It's potentially expensive, but could be useful as a stricter heuristic for deciding whether to rewrite the request body.
        """
        values, errors = await request_body_to_args(self.dependant.body_params, body)
        return False if errors else True



from copy import deepcopy
from typing import Dict, List, Optional

from starlette.responses import HTMLResponse

from fastapitableau.utils import replace_dict_keys


def rewrite_tableau_openapi(
    openapi: Dict, rewrite_paths: Optional[List[str]] = None
) -> Dict:
    openapi = deepcopy(openapi)
    schemas = openapi["components"]["schemas"]

    if rewrite_paths is None:
        rewrite_paths = list(openapi["paths"].keys())

    for path_name in rewrite_paths:

        path = openapi["paths"][path_name]
        request_body = path.get("post").get("requestBody")

        if request_body is None:
            continue  # We skip rewriting functions that just take the whole request

        path_schema = request_body["content"]["application/json"]["schema"]

        if "$ref" not in path_schema.keys():
            # Do some moving around of elements to make the single field appear as a referenced schema

            schema_name = "Body_" + path["post"]["operationId"]
            schema_ref = "#/components/schemas/" + schema_name

            # schema = openapi["components"]["schemas"][schema_name]
            lowercase_title = path_schema["title"].lower()

            schema = {
                "title": schema_name,
                "required": [lowercase_title],
                "type": "object",
                "properties": {lowercase_title: path_schema},
            }

            # Replace components with newly created reference schema
            path["post"]["requestBody"]["content"]["application/json"]["schema"] = {
                "$ref": schema_ref
            }
            openapi["components"]["schemas"][schema_name] = schema

            # Overwrite the path_schema object with the new ref
            path_schema = path["post"]["requestBody"]["content"]["application/json"][
                "schema"
            ]

        schema_ref = path_schema["$ref"]
        schema_name = schema_ref.split("/")[-1]

        tab_schema_name = schema_name + "_tableau"
        tab_schema_ref = schema_ref + "_tableau"

        schema = openapi["components"]["schemas"][schema_name]

        # Generate a list of the expected Tableau labels. If it is in `required`, relabel it there, and add it to the list of new keys.
        new_keys: Dict[str, str] = {}
        for i, key in enumerate(schema["properties"].keys()):
            new_key_name = "_arg" + str(i + 1)
            new_keys[key] = new_key_name
            schema["required"] = [
                x if x != key else new_key_name for x in schema["required"]
            ]
        schema["properties"] = replace_dict_keys(schema["properties"], new_keys)
        tab_schema = {
            "title": tab_schema_name,
            "required": ["script", "data"],
            "type": "object",
            "properties": {
                "script": {"title": "Script", "type": "string", "default": path_name},
                "data": {"$ref": schema_ref},
            },
        }

        # Insert the tableau request schema into schemas and change the path schema ref to point to it.
        schemas[tab_schema_name] = tab_schema
        path_schema["$ref"] = tab_schema_ref

    return openapi


# Vendored and modified from FastAPI.
def get_swagger_ui_html(
    *,
    openapi_url: str,
    title: str,
    swagger_js_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui-bundle.js",
    swagger_css_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@3/swagger-ui.css",
    swagger_favicon_url: str = "https://fastapi.tiangolo.com/img/favicon.png",
    home_url: str,
) -> HTMLResponse:

    html = f"""
<!DOCTYPE html>
    <html>
    <head>
    <link type="text/css" rel="stylesheet" href="{swagger_css_url}">
    <link rel="shortcut icon" href="{swagger_favicon_url}">
    <link rel="stylesheet" type="text/css" href="static/css/styles.css">
    <title>{title}</title>
    </head>
    <body>

    <!-- BEGIN: Insert our header into the documentation -->
    <header class="md-header" data-md-component="header" data-md-state="shadow">
        <nav class="md-header__inner md-grid" aria-label="Header">
            <a href="{home_url}" class="nav-bar">
                <label class="md-header__button md-icon">
                    <!-- back icon -->
                    <svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 122.88 122.88"><title>back</title><path class="cls-1" d="M61.44,0A61.51,61.51,0,1,1,18,18,61.25,61.25,0,0,1,61.44,0Zm5,45.27A7.23,7.23,0,1,0,56.14,35.13L35,56.57a7.24,7.24,0,0,0,0,10.15l20.71,21A7.23,7.23,0,1,0,66.06,77.62l-8.73-8.87,24.86-.15a7.24,7.24,0,1,0-.13-14.47l-24.44.14,8.84-9Z"/></svg>
                </label>
                <div class="md-header__title" data-md-component="header-title">
                    <div class="md-header__ellipsis">
                        <div class="md-header__topic">
                            <span class="md-ellipsis">
                                FastAPI Tableau — {title}
                            </span>
                        </div>
                    </div>
                </div>
            </a>
        </nav>
    </header>
    <!-- END: Insert our header into the documentation -->

    <!-- BEGIN: Small container to make positions consistent between this and other pages -->
    <div class="swagger-container">
    <!-- END: Small container -->

    <div id="swagger-ui">
    </div>

    <!-- BEGIN: Close our small container -->
    </div>
    <!-- END: Close small container -->

    <script src="{swagger_js_url}"></script>
    <!-- `SwaggerUIBundle` is now available on the page -->
    <script>
    const ui = SwaggerUIBundle({{
        url: '{openapi_url}',
    """

    html += """
        dom_id: '#swagger-ui',
        presets: [
        SwaggerUIBundle.presets.apis,
        SwaggerUIBundle.SwaggerUIStandalonePreset
        ],
        layout: "BaseLayout",
        deepLinking: true,
        showExtensions: true,
        showCommonExtensions: true
    })"""

    html += """
    </script>
    </body>
    </html>
    """
    return HTMLResponse(html)



from dataclasses import dataclass
from inspect import Signature, signature
from typing import Any, List, Optional
from urllib.parse import urljoin, urlparse

from fastapitableau.utils import remove_prefix


@dataclass
class ParamInfo:
    name: str
    type: str
    tableau_type: str
    required: bool
    default: Any
    details: str

    def __init__(self, param):
        self.name = param.name
        self.type = param._type_display()
        self.tableau_type = tableau_name_for_python_type(self.type)
        self.required = param.required
        self.default = param.default
        self.details = self._details()

    def _details(self) -> str:
        parts = []
        parts.append("required") if self.required else parts.append("optional")
        if self.default is not None:
            parts.append(f"default = {self.default}")
        return "; ".join(parts).capitalize()


@dataclass
class ReturnInfo:
    type: str
    desc: str

    def __init__(self, route):
        sig = signature(route.dependant.call)
        return_annotation = (
            sig.return_annotation
            if sig.return_annotation is not Signature.empty
            else None
        )
        self.type = remove_prefix(str(return_annotation), "typing.")
        self.desc = (
            route.response_description
            if route.response_description != "Successful Response"
            else ""
        )


@dataclass
class RouteInfo:
    path: str
    usage: str
    summary: str
    description: str
    body_params: List[ParamInfo]
    query_params: List[ParamInfo]
    return_info: ReturnInfo
    name: str

    def __init__(self, route, app_base_url: Optional[str]):
        if app_base_url:
            # Displays the path relative to the Connect domain, for Tableau purposes.
            self.path = urljoin(urlparse(app_base_url).path, route.path.strip("/"))
        else:
            self.path = route.path
        self.description = route.description
        self.summary = route.summary
        self.body_params = [ParamInfo(param) for param in route.dependant.body_params]
        self.query_params = [ParamInfo(param) for param in route.dependant.query_params]
        self.return_info = ReturnInfo(route)
        self.usage = self._tableau_usage_str()
        self.name = route.path.lstrip("/")

    def _tableau_usage_str(self):
        tableau_funcs = {
            "List[bool]": "SCRIPT_BOOL",
            "List[str]": "SCRIPT_STR",
            "List[int]": "SCRIPT_INT",
            "List[float]": "SCRIPT_REAL",
            "": "SCRIPT_STR",
        }

        if self.return_info.type not in tableau_funcs.keys():
            return f"<Warning: Cannot determine Tableau script command for return type '{self.return_info.type}'>"
        else:
            tableau_func = tableau_funcs[self.return_info.type]

        params = ", ".join([p.name for p in self.body_params])

        return f'{tableau_func}("{self.path}", {params})'


def extract_routes_info(app, app_base_url: Optional[str]):
    routes_info = [
        RouteInfo(route, app_base_url)
        for route in app.routes
        if "TableauRoute" in route.__class__.__name__
    ]
    return routes_info


def tableau_name_for_python_type(python_type: str) -> str:
    if "bool" in python_type:
        tableau_type = "Boolean"
    elif "str" in python_type:
        tableau_type = "String"
    elif "int" in python_type:
        tableau_type = "Integer (whole number)"
    elif "float" in python_type:
        tableau_type = "Real (decimal number)"
    elif python_type == "":
        tableau_type = "String"
    else:
        return f"<Warning: No equivalent Tableau type for '{python_type}'>"

    return tableau_type



from typing import Dict

from fastapi import Request
from starlette.types import Receive


def replace_dict_keys(old_dict: Dict, new_keys: Dict[str, str]):
    new_dict: Dict = {}
    for old, new in new_keys.items():
        new_dict[new] = old_dict[old]
    return new_dict


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def calc_app_base_url(request: Request):
    app_base_url = request.headers.get("RStudio-Connect-App-Base-URL")
    if not app_base_url:
        app_base_url = "/"
    elif app_base_url[-1] != "/":
        app_base_url += "/"
    return app_base_url


async def event_from_receive(receive: Receive) -> Dict:
    message_body = b""
    more_body = True
    while more_body:
        message = await receive()
        message_body += message.get("body", b"")
        more_body = message.get("more_body", False)

    event = {
        "type": message.get("type", "http.request"),
        "body": message_body,
        "more_body": more_body,
    }

    return event



import json
from typing import Dict, Tuple

from starlette.types import Receive

from fastapitableau.logger import logger
from fastapitableau.utils import event_from_receive


class TableauExtensionMiddleware:
    """
    Rewrites requests sent to the "/evaluate" endpoint. To do this, it receives the triggering event and uses the body of the request, then passes on the event in an awaitable object.
    """

    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send) -> None:
        logger.debug("Received request with scope: %s", scope, extra={"scope": scope})
        if scope["type"] == "http" and scope["path"] == "/evaluate":
            _scope, _receive = await self.rewrite_scope_path(scope, receive)
            await self.app(_scope, _receive, send)
        else:
            await self.app(scope, receive, send)

    @staticmethod
    async def rewrite_scope_path(scope: Dict, receive: Receive) -> Tuple[Dict, Receive]:
        event = await event_from_receive(receive)

        logger.debug("Received event: %s", event, extra={"scope": scope})

        try:
            body = json.loads(event["body"])
        except Exception as e:
            print("Failed to parse event body as JSON: {event}")
            print(e)
            raise e

        target_path = body["script"]
        if target_path[0] != "/":
            target_path = "/" + target_path

        logger.debug(
            "Rewriting path for Tableau request: '%s'",
            target_path,
            extra={"scope": scope},
        )

        scope["path"] = target_path
        scope["raw_path"] = bytes(target_path, encoding="utf-8")

        async def _receive():
            return event

        return scope, _receive



from os import environ, getcwd
from typing import List, Optional
from urllib.parse import urlparse

import requests

from fastapitableau.logger import logger


def check_rstudio_connect() -> bool:
    """Returns True if running in RStudio Connect"""
    checks = [
        # The first check is only valid for recent versions.
        environ.get("RSTUDIO_PRODUCT") == "CONNECT",
        # We still perform checks for evidence we're running on older Connect versions so we can show warning messages about upgrading.
        "RSTUDIO_CONNECT_HASTE" in environ.keys(),
        getcwd() == "/opt/rstudio-connect/mnt/app",
        environ.get("LOGNAME") == "rstudio-connect",
        environ.get("TMPDIR") == "/opt/rstudio-connect/mnt/tmp",
    ]
    rstudio_connect = any(checks)
    if rstudio_connect:
        logger.debug("We seem to be running in RStudio Connect")
    else:
        logger.debug("We don't seem to be running in RStudio Connect")
    return any(checks)


def warning_message() -> Optional[str]:  # noqa: C901
    """
    Generate a string of warning messages based on information gathered about our environment in RStudio Connect.
    """
    if not check_rstudio_connect():
        return None

    message_list: List[str] = []

    connect_server = environ.get("CONNECT_SERVER")

    if connect_server is None:
        logger.warning("Could not find CONNECT_SERVER environment variable")
        message_list.append(
            "\n### The environment variable `CONNECT_SERVER` is not defined\n"
            "\n"
            "Possible Solutions:\n"
            "\n"
            "- Have your system administrator confirm `Applications.DefaultServerEnv` is enabled and that `Server.Address` has been defined within the `rstudio-connect.gcfg` file on the RStudio Connect server.,\n"
            "- Use the application settings for your content within the RStudio Connect dashboard to define the `CONNECT_SERVER` environment variable. It should be set to a reachable https or http address for the server.\n"
        )

    if urlparse(connect_server).scheme is None:
        logger.warning(
            "The CONNECT_SERVER environment variable doesn't specify https:// or http://"
        )
        message_list.append(
            f"### Environment Variable `CONNECT_SERVER` (value = `{connect_server}` ) does not specify the protocol (`https://` or `http://`)\n"
            "\n"
            "Possible Solutions:\n"
            "\n"
            "- Have your system administrator confirm that `Server.Address` has been configured with the proper format within the `rstudio-connect.gcfg` file on the RStudio Connect server.\n"
            "- Use the application settings for your content within the RStudio Connect dashboard to define the CONNECT_SERVER` environment variable with the proper protocol.\n"
        )

    connect_api_key = environ.get("CONNECT_API_KEY")

    if connect_api_key is None:
        logger.warning("Could not find CONNECT_API_KEY environment variable")
        message_list.append(
            "### The environment variable `CONNECT_API_KEY` is not defined\n"
            "\n"
            "Possible Solutions:\n"
            "\n"
            "- Have your administrator enable the `Applications.DefaultAPIKeyEnv` setting within the `rstudio-connect.gcfg` file on the RStudio Connect server.\n"
            "- Create an API key yourself and use the application settings for your content within the RStudio Connect dashboard to set the the `CONNECT_API_KEY` variable to its value.\n"
        )

    if len(message_list) != 0:
        messages = "\n\n---\n\n".join(message_list)
        return messages

    try:
        # Call RStudio Connect API to get server settings
        use_http = environ.get("FASTAPITABLEAU_USE_HTTP", "False").title() == "True"
        if use_http:
            connect_server = urlparse(connect_server)._replace(scheme="http").geturl()  # type: ignore[arg-type]
        settings_url = str(connect_server) + "__api__/server_settings"
        headers = {"Authorization": "Key " + str(connect_api_key)}
        response = requests.get(settings_url, headers=headers, verify=not use_http)
    except Exception as e:
        logger.warning(
            "Unable to access RStudio Connect settings API due to error: %s", e
        )
        message_list.append(
            f"### API request to {connect_server} has failed with error:\n"
            f"{e}"
            "\n"
            "\nPossible Solutions:"
            "\n"
            "\n- If you have specified an API key, confirm it is valid."
            "\n- Confirm there is connectivity between the server itself and the address assigned to it: "
            f"{connect_server}"
            "."
            "\n- If using HTTPS with self-signed certificates, you need to tell Python about the location of your certificate file. In RStudio Connect, set the `REQUESTS_CA_BUNDLE` environment variable to the path to your certificate file, in the Vars tab of the Content Settings Panel. "
            "Note that applications on RStudio Connect cannot access `/etc` directory, so the certificate must be located elsewhere, such as `/opt/python`."
        )
    else:
        # Only execute if response exists
        if response.status_code != 200:
            logger.warning(
                "Unable to access RStudio Connect settings API with status code: %s",
                response.status_code,
            )
            message_list.append(
                f"### API request to {connect_server} has failed with error: {response.text}\n"
                "\n"
                "Possible Solutions:\n"
                "\n"
                "- If you have specified an API_KEY, confirm it is valid.\n"
                f"- Confirm the server can be reached at {connect_server}.\n"
            )
        else:
            server_settings = response.json()
            if "tableau_integration_enabled" not in server_settings.keys():
                logger.warning(
                    "The Tableau integration feature flag is not available on this RStudio Connect server"
                )
                message_list.append(
                    "### Tableau Integration Feature Flag is not available on the RStudio Connect server.\n"
                    "\n"
                    "Possible Solution:\n"
                    "\n"
                    "- Please upgrade to the latest version of RStudio Connect.\n"
                )
            elif server_settings["tableau_integration_enabled"] is False:
                logger.warning(
                    "Tableau integration is disabled on this RStudio Connect server"
                )
                message_list.append(
                    "Tableau Integration has been disabled on the RStudio Connect server\n"
                    "\n"
                    "Possible Solution:\n"
                    "\n"
                    "- Please ask your administrator to set `TableauIntegration.Enabled` = `true` within `rstudio-connect.gcfg` file on the RStudio Connect server.\n"
                )

    if len(message_list) != 0:
        print(message_list)
        messages = "\n\n---\n\n".join(message_list)
        return messages
    else:
        return None



import os
from urllib.parse import urlparse

from commonmark import commonmark  # type: ignore[import]
from fastapi import Request
from fastapi.routing import APIRouter
from pkg_resources import resource_filename
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates

from fastapitableau import rstudio_connect
from fastapitableau.logger import logger
from fastapitableau.openapi import get_swagger_ui_html
from fastapitableau.user_guide import extract_routes_info
from fastapitableau.utils import calc_app_base_url


def markdown_filter(text):
    return commonmark(text)


templates_directory = resource_filename("fastapitableau", "templates")

statics = StaticFiles(packages=["fastapitableau"])
jinja_templates = Jinja2Templates(directory=templates_directory)
jinja_templates.env.filters["markdown"] = markdown_filter

built_in_pages = APIRouter()


@built_in_pages.get("/", include_in_schema=False)
async def home(request: Request):
    context = {
        "request": request,
        "warning_message": rstudio_connect.warning_message(),
        "app_base_url": calc_app_base_url(request),
    }
    return jinja_templates.TemplateResponse("index.html", context=context)


@built_in_pages.get("/setup_tableau", include_in_schema=False)
async def setup(request: Request):
    logger.debug("Generating Tableau setup instructions")
    server = os.getenv("CONNECT_SERVER", "")
    parts = urlparse(server)
    # <scheme>://<netloc>/<path>;<params>?<query>#<fragment>
    server_domain = (
        "<The URL for your RStudio Connect server (e.g. https://connect.example.com)>"
    )
    server_port = "<The port on which your RStudio Connect is listening>"
    netloc = parts.netloc.split(":")
    if len(netloc) > 1:
        server_domain = parts.scheme + "://" + netloc[0]
        server_port = netloc[1]

    context = {
        "request": request,
        "warning_message": rstudio_connect.warning_message(),
        "server_domain": server_domain,
        "server_port": server_port,
        "app_base_url": calc_app_base_url(request),
    }
    return jinja_templates.TemplateResponse("setup_tableau.html", context=context)


@built_in_pages.get("/tableau_usage", include_in_schema=False)
async def tableau_usage(request: Request):
    logger.debug("Generating Tableau usage instructions")
    routes_info = extract_routes_info(
        app=request.app,
        app_base_url=calc_app_base_url(request),
    )
    context = {
        "request": request,
        "warning_message": rstudio_connect.warning_message(),
        "routes_info": routes_info,
        "app_base_url": calc_app_base_url(request),
    }
    return jinja_templates.TemplateResponse("tableau_usage.html", context=context)


# We're bypassing the built-in generation of the docs so we can show two different
# versions of the OpenAPI docs: the default and then one reconfigured for Tableau.
@built_in_pages.get("/docs_standard", include_in_schema=False)
async def docs_openAPI(request: Request):
    root_path = request.scope.get("root_path", "").rstrip("/")
    openapi_url = root_path + "/openapi.json"
    request.app.use_tableau_api_schema = False
    app_base_url = calc_app_base_url(request)

    return get_swagger_ui_html(
        openapi_url=openapi_url,
        title="OpenAPI: Standard Web Requests",
        home_url=app_base_url,
    )


@built_in_pages.get("/docs_tableau", include_in_schema=False)
async def docs_tableau_openAPI(request: Request):
    root_path = request.scope.get("root_path", "").rstrip("/")
    openapi_url = root_path + "/openapi.json"
    app_base_url = calc_app_base_url(request)

    request.app.use_tableau_api_schema = True
    return get_swagger_ui_html(
        openapi_url=openapi_url,
        title="OpenAPI: Tableau-Style Requests",
        home_url=app_base_url,
    )



import logging
import os
import sys
from typing import Optional

from starlette.datastructures import Headers
from starlette.types import Scope


class ScopeAdapter(logging.LoggerAdapter):
    """
    This adapter will prepend the correlation ID of a scope dict passed in via
    the "scope" keyword argument.
    """

    def process(self, msg, kwargs):
        extra = kwargs.get("extra", {})
        if "scope" not in extra.keys():
            return msg, kwargs
        else:
            correlation_id = get_correlation_id(extra["scope"])
            kwargs["extra"]["correlation_id"] = correlation_id
            prefix = f"[{correlation_id}] " if correlation_id is not None else ""
            return prefix + msg, kwargs


def get_correlation_id(scope: Scope = None, headers: Headers = None) -> Optional[str]:
    headers = Headers(scope=scope, headers=headers)
    correlation_id_keys = ["x-rs-correlation-id", "x-correlation-id"]
    for key in correlation_id_keys:
        if key in headers.keys():
            return headers[key]
    return None


base_logger = logging.getLogger("fastapitableau")
base_logger.setLevel(
    getattr(logging, os.environ.get("FASTAPITABLEAU_LOG_LEVEL", "INFO").upper())
)
logging.basicConfig(format="%(message)s", stream=sys.stdout)

# FORMAT = '%(correlation_id)-20s %(message)s'
# logging.basicConfig(format=FORMAT, stream=sys.stdout)

logger = ScopeAdapter(base_logger, {})



import json
from typing import Any, Dict

from fastapi import APIRouter, FastAPI, Request
from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException

from fastapitableau.exception_handlers import (
    tableau_general_exception_handler,
    tableau_http_exception_handler,
    tableau_request_validation_exception_handler,
)
from fastapitableau.logger import logger
from fastapitableau.middleware import TableauExtensionMiddleware
from fastapitableau.openapi import rewrite_tableau_openapi
from fastapitableau.pages import built_in_pages, statics
from fastapitableau.routing import TableauRoute


class FastAPITableau(FastAPI):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_middleware(TableauExtensionMiddleware)
        self.router.route_class = TableauRoute
        self.mount("/static", statics, name="static")
        self.include_router(built_in_pages)
        self.include_router(info_router)
        self.docs_url = None  # We will be serving up our own docs
        self.use_tableau_api_schema = False
        if not self.description:
            self.description = "Description not provided"

        # Add exception handlers
        self.exception_handlers[HTTPException] = tableau_http_exception_handler
        self.exception_handlers[
            RequestValidationError
        ] = tableau_request_validation_exception_handler
        self.exception_handlers[Exception] = tableau_general_exception_handler
        self.middleware_stack = self.build_middleware_stack()

    def openapi(self) -> Dict[str, Any]:
        orig_desc = self.description
        self.openapi_schema = None
        if self.use_tableau_api_schema:
            logger.debug("Generating OpenAPI schema for Tableau requests...")
            self.description = (
                orig_desc
                + '<br><br>*NOTE: This page\'s example API requests are formatted like the requests that Tableau will send. They are different from standard web requests, which are documented under "Test Standard Web Requests".*'
            )
            schema = super().openapi()
            tableau_paths = [
                route.path for route in self.routes if isinstance(route, TableauRoute)
            ]
            self.openapi_schema = rewrite_tableau_openapi(schema, tableau_paths)
            logger.debug("Schema JSON: %s", json.dumps(self.openapi_schema))
        else:
            logger.debug("Generating OpenAPI schema for standard requests...")
            self.description = (
                orig_desc
                + '<br><br>*NOTE: This page\'s example API requests are formatted like standard web requests. They are different from the requests that Tableau will send, which are documented under "Test Tableau-Style Requests".*'
            )
            self.openapi_schema = super().openapi()
            logger.debug("Schema JSON: %s", json.dumps(self.openapi_schema))

        self.description = orig_desc
        return self.openapi_schema


info_router = APIRouter()


@info_router.get("/info", include_in_schema=False)
def info(request: Request):
    logger.debug("Responding to '/info' request", extra={"scope": request.scope})
    return {
        "description": "FastAPITableau API",
        # "creation_time": "0",
        # "state_path": "e:\\dev\\server\\server\\server",
        "server_version": "0.0.1",
        "name": request.app.title,
        "versions": {"v1": {"features": {}}},
    }
